;;; Copyright (C) 2020 Wolfgang Corcoran-Mathe

;;; Permission is hereby granted, free of charge, to any person obtaining a
;;; copy of this software and associated documentation files (the
;;; "Software"), to deal in the Software without restriction, including
;;; without limitation the rights to use, copy, modify, merge, publish,
;;; distribute, sublicense, and/or sell copies of the Software, and to
;;; permit persons to whom the Software is furnished to do so, subject to
;;; the following conditions:

;;; The above copyright notice and this permission notice shall be included
;;; in all copies or substantial portions of the Software.

;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
;;; OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
;;; IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
;;; CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
;;; TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
;;; SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

(import (except scheme list->string string->list string-ref
                       string-length)
        (chicken base)
        (chicken condition)
        (chicken fixnum)
        (chicken random)
        (chicken syntax)
        (srfi 1)
        (only (srfi 133) vector-map vector-every vector-index-right)
        (only utf8 list->string string->list string-ref string-length)
        test
        test-generative
        (srfi 196))

(define random-int pseudo-random-integer)

(define (generator->list g)
  (let ((v (g)))
    (if (eof-object? v)
        '()
        (cons v (generator->list g)))))

;;;; Utility

(define always (constantly #t))
(define never (constantly #f))

(define (%range-empty? r) (zero? (range-length r)))

;; Apply f k times recursively to seed.
(define (iterate k f seed)
  (unless (and (integer? k) (>= k 0))
    (error 'iterate "invalid count" k))
  (do ((k k (- k 1))
       (seed seed (f seed)))
      ((= k 0) seed)))

;;;; Test ranges

(define test-num-range (numeric-range 10 30))

(define test-num-seq (iota 20 10))

(define test-empty-range (numeric-range 0 0))

(define (square x) (* x x))

;; Produces the range {#f, #t}.
(define test-bool-range
  (range 2 (lambda (n) (not (zero? n)))))

(define test-length-bound 64)

;; Returns a vector of random length containing random fixnums.
(define (random-vector)
  (let ((len (random-int test-length-bound)))
    (do ((i 0 (+ i 1))
         (vec (make-vector len)))
        ((= i len) vec)
      (vector-set! vec i (random-int most-positive-fixnum)))))

;; Returns a range of random length containing random fixnums.
(define (random-range) (vector-range (random-vector)))

;; Use test-generative to bind a list of names to random ranges
;; over a bunch of body expressions.
(define-syntax test/random-ranges
  (ir-macro-transformer
   (lambda (exp _inj _same?)
     `(test-generative ,(map (lambda (v) (list v 'random-range))
                             (cadr exp))
        ,@(cddr exp)))))

;; String of random characters in the BMP.
(define (random-string)
  (list->string
   (list-tabulate (random-int test-length-bound)
                  (lambda (junk)
                    (integer->char (random-int #xffff))))))

;; True if exp raises a condition with kind (exn bounds ...).
(define-syntax raises-bounds-exn
  (syntax-rules ()
    ((raises-bounds-exn exp)
     (condition-case exp
       ((exn bounds) #t)
       (var () #f)))))

;; True if exp raises a condition with kind (exn type ...).
(define-syntax raises-type-exn
  (syntax-rules ()
    ((raises-type-exn exp)
     (condition-case exp
       ((exn type) #t)
       (var () #f)))))

;;;; Conversion

;;; These see a lot of use in the following tests, so let's check
;;; them first.

(test-group "Conversion"
  (test-group "range->list"
    (test '() (range->list test-empty-range))
    (test '(#f #t) (range->list test-bool-range))
    (test test-num-seq (range->list test-num-range))
    )

  (test-group "generator->list"
    (test '() (generator->list (range->generator test-empty-range)))
    (test '(#f #t) (generator->list (range->generator test-bool-range)))
    (test test-num-seq (generator->list (range->generator test-num-range)))
    )

  (test-group "range->vector"
    (test '#() (range->vector test-empty-range))
    (test '#(#f #t) (range->vector test-bool-range))
    (test test-num-seq (vector->list (range->vector test-num-range)))
    )

  (test-group "range->string"
    (test "" (range->string test-empty-range))
    (let ((s "0123456789"))
      (test s (range->string (string-range s))))
    )
  )

(test-group "Constructors"
  (test-group "numeric-range"
    (test #t (%range-empty? (numeric-range 1 1)))
    (test (iota 4 -5) (range->list (numeric-range -5 -1)))
    (test (iota 6 1 -1) (range->list (numeric-range 1 -5 -1)))
    (test (iota 4 4/3) (range->list (numeric-range 4/3 16/3)))
    (test (iota 3 0 4) (range->list (numeric-range 0 9 4)))
    (test #t (%range-empty? (numeric-range 0 10 -1)))
    (test #t (%range-empty? (numeric-range 0 -10)))
    (test (iota 4 5 -1) (range->list (numeric-range 5 1 -1)))
    (test (iota 4 -2) (range->list (numeric-range -2 2)))
    (test (iota 4 2 -1) (range->list (numeric-range 2 -2 -1)))
    (test (iota 4 -4 -1) (range->list (numeric-range -4 -8 -1)))
    (test (iota 5 -1 -2/3) (range->list (numeric-range -1 -4 -2/3)))

    (test-generative ((start (lambda () (random-int 65536)))
                      (step (lambda () (max 1 (random-int 1024)))))
      (let ((end (+ start (random-int 32768))))
        (test (unfold (lambda (k) (>= k end))
                      values
                      (lambda (k) (+ k step))
                      start)
              (range->list (numeric-range start end step)))))
    )

  (test-group "iota-range"
    (test #t (range=? eqv? (iota-range 10 0 0) (range 10 (lambda (_i) 0))))
    (test #t (%range-empty? (iota-range 0)))
    (test (iota 10) (range->list (iota-range 10)))
    (test (iota 10) (range->list (iota-range 10 0)))
    (test (iota 10) (range->list (iota-range 10 0 1)))
    (test (iota 10 10 2) (range->list (iota-range 10 10 2)))
    (test (iota 10 0 -1) (range->list (iota-range 10 0 -1)))
    (test (iota 10 5 -2) (range->list (iota-range 10 5 -2)))
    (test (iota 10 1/2) (range->list (iota-range 10 1/2)))

    (test-generative ((len (lambda () (random-int test-length-bound)))
                      (start (lambda () (random-int 65536)))
                      (step (lambda () (random-int 1024))))
      (test (iota len start step)
            (range->list (iota-range len start step))))
    )

  (test-group "vector-range"
    (test-generative ((vec random-vector))
      (test (vector-length vec) (range-length (vector-range vec)))
      (test vec (range->vector (vector-range vec)))
      )
    )

  (test-group "string-range"
    (test-generative ((s random-string))
      (test (string-length s) (range-length (string-range s)))
      (test (string->list s) (range->list (string-range s)))
      )
    )

  (test-group "range-append"
    (test 0 (range-length (range-append)))
    (test 0 (range-length (range-append (iota-range 0))))
    (test 0 (range-length (range-append (iota-range 0) (iota-range 0))))
    (test 0 (range-length (range-append (iota-range 0)
                                        (iota-range 0)
                                        (iota-range 0))))

    (test/random-ranges (r1 r2 r3)
      (test-assert (range=? = r1 (range-append r1)))
      (test-assert
       (range=? = r1 (range-append (iota-range 0) r1 (iota-range 0))))
      (test (append (range->list r1) (range->list r2))
            (range->list (range-append r1 r2)))
      (test (append (range->list r3) (range->list r2) (range->list r1))
            (range->list (range-append r3 r2 r1)))
      )

    (test-assert
     (range=? =
              (range-append (numeric-range 10 15)
                            (numeric-range 15 20)
                            (numeric-range 20 25)
                            (numeric-range 25 30))
              (numeric-range 10 30)))

    (test-assert (raises-type-exn (range-append #t)))
    (test-assert (raises-type-exn (range-append (iota-range 0) #t)))
    (test-assert
     (raises-type-exn (range-append (iota-range 0) (iota-range 0) 10)))
    )

  (test-group "range-reverse"
    (test 0 (range-length (range-reverse (iota-range 0))))

    (test/random-ranges (r)
      (test-generative ((k (lambda () (random-int 20))))
        (test-assert
         (let* ((r* (iterate k range-reverse r))
                (lis (range->list r*)))
           (if (even? k)
               (equal? lis (range->list r))
               (equal? lis (reverse (range->list r)))))))
      )
    (test-assert (raises-type-exn (range-reverse #t)))
    )
  )

;;;; Predicates

(test-group "Predicates"
  (test-group "range?"
    (test #t (range? test-empty-range))
    (test #t (range? test-num-range))
    (test #f (range? #t))
    )

  (test-group "range=?"
    (test #t (range=? eqv? test-empty-range (iota-range 0)))
    (test #t (range=? = (numeric-range 0 0) (numeric-range 5 5)))
    (test #f (range=? = (numeric-range 0 0) test-num-range))
    (test #t (range=? = test-num-range test-num-range))
    (test #t (range=? = test-num-range (numeric-range 10 30)))
    (test #f (range=? = test-num-range (numeric-range 10 20)))
    (test #t (range=? eqv? test-bool-range (vector-range #(#f #t))))
    (test #f (range=? eqv? test-bool-range (vector-range #(#t #f))))
    (test #t (range=? eqv?
                      test-empty-range
                      (iota-range 0)
                      (numeric-range 0 0)))
    (test #f (range=? =
                      test-empty-range
                      (iota-range 1)
                      (numeric-range 0 0)))
    (test #t (range=? eqv?
                      (range 5 values)
                      (iota-range 5)
                      (numeric-range 0 5)))
    )
  )

;;;; Accessors

(test-group "Accessors"
  (test-group "range-length"
    (test 0 (range-length (iota-range 0)))
    (test/random-ranges (r)
      (test (length (range->list r)) (range-length r)))
    )

  (test-group "range-first"
    (test #f (range-first test-bool-range))
    (test 0 (range-first (iota-range 1)))
    (test-assert (raises-bounds-exn (range-first (iota-range 0))))
    (test-generative ((start (lambda () (random-int 65536)))
                      (step (lambda () (random-int 65536))))
      (let* ((len 10)
             (r (iota-range len start step)))
        (test start (range-first r))
        ))
    )

  (test-group "range-last"
    (test #t (range-last test-bool-range))
    (test 0 (range-last (iota-range 1)))
    (test-assert (raises-bounds-exn (range-last (iota-range 0))))
    (test-generative ((start (lambda () (random-int 65536)))
                      (step (lambda () (random-int 65536))))
      (let* ((len 10)
             (r (iota-range len start step)))
        (test (+ start (* step 9)) (range-last r))
        ))
    )
  )

;;;; Iteration

(test-group "Iteration"
  (test-group "range-split-at"
    (test '(() ())
          (let-values ((rs (range-split-at (iota-range 0) 0)))
            (map range->list rs)))

    (test/random-ranges (r)
      (let ((len (range-length r)))
        (test-assert
         (let-values (((ra rb) (range-split-at r 0)))
           (and (zero? (range-length ra)) (range=? = r rb))))

        (test-assert
         (let-values (((ra rb) (range-split-at r len)))
           (and (range=? = r ra) (zero? (range-length rb)))))

        (test-assert
         (let-values (((r-ones r*)
                       (range-split-at
                        (range-append (range 5 (constantly 'z)) r)
                        5)))
           (and (equal? '(z z z z z) (range->list r-ones))
                (range=? = r r*))))

        (test-assert
         (let-values (((r* r-ones)
                       (range-split-at
                        (range-append r (range 5 (constantly 'z)))
                        len)))
           (and (range=? = r r*)
                (equal? '(z z z z z) (range->list r-ones)))))

        (test-assert (raises-bounds-exn (range-split-at r (+ len 1))))
        ))

    (test-assert (raises-type-exn (range-split-at #t 0)))
    (test-assert (raises-type-exn (range-split-at (iota-range 0) #t)))
    )

  (test-group "subrange"
    (test '() (range->list (subrange (iota-range 0) 0 0)))

    (test/random-ranges (r)
      (let ((len (range-length r)))
        (test '() (range->list (subrange r 0 0)))
        (test-assert (range=? = r (subrange r 0 len)))

        (let ((r* (range-append (range 5 (constantly 'z)) r)))
          (test '(z z z z z) (range->list (subrange r* 0 5)))
          (test-assert (range=? = r (subrange r* 5 (+ len 5))))
          (test (append '(z z) (range->list r))
                (range->list (subrange r* 3 (+ len 5)))))

        (test-assert (raises-bounds-exn (subrange r -1 len)))
        (test-assert (raises-bounds-exn (subrange r 0 (+ len 1))))
        (test-assert (raises-bounds-exn (subrange r len 0)))
        ))

    (test-assert (raises-type-exn (subrange #t 0 0)))
    (test-assert (raises-type-exn (subrange (iota-range 0) 2.3 0)))
    (test-assert (raises-type-exn (subrange (iota-range 0) 0 2.3)))
  )

  (test-group "range-take"
    (test '() (range->list (range-take (iota-range 0) 0)))

    (test/random-ranges (r)
      (let ((len (range-length r)))
        (test '() (range->list (range-take r 0)))
        (test-assert (range=? = r (range-take r len)))

        (test
         (append (range->list r) '(z z))
         (let ((r* (range-append r (range 5 (constantly 'z)))))
           (range->list (range-take r* (+ len 2)))))

        (test
         '(z z z z z)
         (let ((r* (range-append (range 5 (constantly 'z)) r)))
           (range->list (range-take r* 5))))

        (test-assert (raises-bounds-exn (range-take r (+ len 1))))
        ))

    (test-assert (raises-type-exn (range-take #t 0)))
    (test-assert (raises-type-exn (range-take (iota-range 0) 2.3)))
    )

  (test-group "range-take-right"
    (test '() (range->list (range-take-right (iota-range 0) 0)))

    (test/random-ranges (r)
      (let ((len (range-length r)))
        (test '() (range->list (range-take-right r 0)))
        (test-assert (range=? = r (range-take-right r len)))

        (test
         (append '(z z) (range->list r))
         (let ((r* (range-append (range 5 (constantly 'z)) r)))
           (range->list (range-take-right r* (+ len 2)))))

        (test
         '(z z z z z)
         (let ((r* (range-append r (range 5 (constantly 'z)))))
           (range->list (range-take-right r* 5))))

        (test-assert (raises-bounds-exn (range-take-right r (+ len 1))))
        ))

    (test-assert (raises-type-exn (range-take-right #t 0)))
    (test-assert (raises-type-exn (range-take-right (iota-range 0) 2.3)))
    )

  (test-group "range-drop"
    (test '() (range->list (range-drop (iota-range 0) 0)))

    (test/random-ranges (r)
      (let ((len (range-length r)))
        (test '() (range->list (range-drop r len)))
        (test-assert (range=? = r (range-drop r 0)))

        (test
         (append '(z z) (range->list r))
         (let ((r* (range-append (range 5 (constantly 'z)) r)))
           (range->list (range-drop r* 3))))

        (test
         '(z z z z z)
         (let ((r* (range-append r (range 5 (constantly 'z)))))
           (range->list (range-drop r* len))))

        (test-assert (raises-bounds-exn (range-drop r (+ len 1))))
        ))

    (test-assert (raises-type-exn (range-drop #t 0)))
    (test-assert (raises-type-exn (range-drop (iota-range 0) 2.3)))
    )

  (test-group "range-drop-right"
    (test '() (range->list (range-drop-right (iota-range 0) 0)))

    (test/random-ranges (r)
      (let ((len (range-length r)))
        (test '() (range->list (range-drop-right r len)))
        (test-assert (range=? = r (range-drop-right r 0)))

        (test
         (append (range->list r) '(z z))
         (let ((r* (range-append r (range 5 (constantly 'z)))))
           (range->list (range-drop-right r* 3))))

        (test
         '(z z z z z)
         (let ((r* (range-append (range 5 (constantly 'z)) r)))
           (range->list (range-drop-right r* len))))

        (test-assert (raises-bounds-exn (range-drop-right r (+ len 1))))
        ))

    (test-assert (raises-type-exn (range-drop-right #t 0)))
    (test-assert (raises-type-exn (range-drop-right (iota-range 0) 2.3)))
    )

  (test-group "range-segment"
    (test '() (range-segment (iota-range 0) 1))

    (test/random-ranges (r)
      (test-assert
       (or (%range-empty? r)
           (range=? = r (car (range-segment r (range-length r))))))

      (test (map list (range->list r))
            (map range->list (range-segment r 1)))
      )

    (test-assert (raises-type-exn (range-segment #t 1)))
    (test-assert (raises-type-exn (range-segment (iota-range 0) 0.2)))
    (test-error  (range-segment (iota-range 0) 0))
    )

  (test-group "range-count"
    (test 0 (range-count always (iota-range 0)))

    (test/random-ranges (r1 r2)
      (test 0 (range-count never r1))
      (test (range-length r1) (range-count always r1))
      (test 3 (range-count symbol?
                           (range-append r1 (range 3 (constantly 'z)))))
      (test (range-length r1)
            (range-count (lambda (x y) (zero? (+ x y)))
                         r1
                         (range-map - r1)))
      (test 0 (range-count never r1 r2))
      (test (min (range-length r1) (range-length r2))
            (range-count always r1 r2))
      (test-assert
       (<= (range-count (lambda (_x _y z) (symbol? z))
                        r1
                        r2
                        (range 3 (constantly 'z)))
           3))
      )

    (test-assert (raises-type-exn (range-count #t (iota-range 0))))
    (test-assert (raises-type-exn (range-count always 0.2)))
    (test-assert
     (raises-type-exn (range-count always (iota-range 0) 0.2)))
    )

  (test-group "range-any"
    (test #f (range-any always (iota-range 0)))

    (test/random-ranges (r1 r2)
      (test #f (range-any never r1))
      (test #f (range-any never r1 r2))
      (test #f (range-any values r1 r2 (iota-range 0)))
      )
    (test-assert (raises-type-exn (range-any #t (iota-range 0))))
    (test-assert (raises-type-exn (range-any always 0.2)))
    (test-assert (raises-type-exn (range-any always (iota-range 0) 0.2)))
    )

  (test-group "range-every"
    (test #t (range-every always (iota-range 0)))

    (test/random-ranges (r1 r2)
      (test (%range-empty? r1) (range-every never r1))
      (test (any %range-empty? (list r1 r2)) (range-every never r1 r2))
      (test #t (range-every (constantly #t) r1 r2 (iota-range 0)))
      (when (> (range-length r1) 0)
        (test (range-last r1) (range-every values r1)))
      )
    (test-assert (raises-type-exn (range-every #t (iota-range 0))))
    (test-assert (raises-type-exn (range-every always 0.2)))
    (test-assert (raises-type-exn (range-every always (iota-range 0) 0.2)))
    )

  ;;; map, filter-map, & for-each

  (test-group "range-map"
    (test '() (range->list (range-map add1 (iota-range 0))))

    (test-assert
     (range=? eqv?
              (range-map add1 test-num-range)
              (numeric-range 11 31)))
    (test (map square test-num-seq)
          (range->list (range-map square test-num-range)))
    (test-assert
     (range=? eqv? (range-map + test-num-range test-num-range)
                   (numeric-range 20 60 2)))

    (test/random-ranges (r1 r2)
      (test-assert (range=? = r1 (range-map values r1)))
      (test-assert (range-every (lambda (x) (eqv? x 'z))
                                (range-map (constantly 'z) r1)))
      (test (range-length r1)
            (range-length (range-map (constantly 'z) r1)))
      (test (map add1 (range->list r1))
            (range->list (range-map add1 r1)))
      ;; map & range-map have the same behavior on short lists/ranges.
      (test (map + (range->list r1) (range->list r2))
            (range->list (range-map + r1 r2)))
      (test 0 (range-length
               (range-map (lambda (x y z) x) r1 r2 (iota-range 0))))
      )

    (test-assert (raises-type-exn (range-map #t (iota-range 0))))
    (test-assert (raises-type-exn (range-map values #f)))
    (test-assert (raises-type-exn (range-map values (iota-range 0) 2)))
    )

  (test-group "range-map->list"
    (test '() (range-map->list values (iota-range 0)))

    (test (iota 20 11) (range-map->list add1 test-num-range))
    (test (map square test-num-seq)
          (range-map->list square test-num-range))
    (test (iota 20 20 2)
          (range-map->list + test-num-range test-num-range))

    (test/random-ranges (r1 r2)
      (test (range->list r1) (range-map->list values r1))
      (test-assert (every (lambda (x) (eqv? x 'z))
                          (range-map->list (constantly 'z) r1)))
      (test (range-length r1)
            (length (range-map->list (constantly 'z) r1)))
      (test (map add1 (range->list r1)) (range-map->list add1 r1))
      ;; map & range-map->list have the same behavior on
      ;; short lists/ranges.
      (test (map + (range->list r1) (range->list r2))
            (range-map->list + r1 r2))
      (test '()
            (range-map->list (lambda (x y z) x) r1 r2 (iota-range 0)))
      )
    (test-assert (raises-type-exn (range-map->list #t (iota-range 0))))
    (test-assert (raises-type-exn (range-map->list values #f)))
    (test-assert
     (raises-type-exn (range-map->list values (iota-range 0) 2)))
    )

  (test-group "range-map->vector"
    (test '#() (range-map->vector values (iota-range 0)))

    (test (list->vector (iota 20 11))
          (range-map->vector add1 test-num-range))
    (test (list->vector (map square test-num-seq))
          (range-map->vector square test-num-range))
    (test (list->vector (iota 20 20 2))
          (range-map->vector + test-num-range test-num-range))

    (test/random-ranges (r1 r2)
      (test (range->vector r1) (range-map->vector values r1))
      (test-assert
       (vector-every (lambda (x) (eqv? x 'z))
                     (range-map->vector (constantly 'z) r1)))
      (test (range-length r1)
            (vector-length (range-map->vector (constantly 'z) r1)))
      (test (vector-map add1 (range->vector r1))
            (range-map->vector add1 r1))
      ;; vector-map & range-map->vector have the same behavior on
      ;; short lists/ranges.
      (test (vector-map + (range->vector r1) (range->vector r2))
            (range-map->vector + r1 r2))
      (test '#()
            (range-map->vector (lambda (x y z) x) r1 r2 (iota-range 0)))
      )
    (test-assert (raises-type-exn (range-map->vector #t (iota-range 0))))
    (test-assert (raises-type-exn (range-map->vector values #f)))
    (test-assert
     (raises-type-exn (range-map->vector values (iota-range 0) 2)))
    )

  (test-group "range-for-each"
    (test 'z
          (let ((foo 'z))
            (range-for-each (lambda (x) (set! foo x))
                            (iota-range 0))
            foo))

    (test/random-ranges (r1 r2)
      (test (reverse (range->list r1))
            (let ((ks '()))
              (range-for-each (lambda (k) (set! ks (cons k ks))) r1)
              ks))
      (test (apply min (map range-length (list r1 r2)))
            (let ((k 0))
              (range-for-each (lambda (_m _n) (set! k (+ k 1))) r1 r2)
              k))
      )
    (test-assert (raises-type-exn (range-for-each #t (iota-range 0))))
    (test-assert (raises-type-exn (range-for-each values 0.3)))
    (test-assert
     (raises-type-exn (range-for-each values (iota-range 0) 'z)))
    )

  (test-group "range-filter-map"
    (test 0 (range-length
             (range-filter-map (constantly #t) (iota-range 0))))

    (test/random-ranges (r1 r2)
      (test 0 (range-length (range-filter-map (constantly #f) r1)))
      (test-assert (range=? = r1 (range-filter-map values r1)))
      (test-assert (range-every (lambda (x) (eqv? x 'z))
                                (range-filter-map (constantly 'z) r1)))
      (test (range-length r1)
            (range-length (range-filter-map (constantly 'z) r1)))
      (test (map add1 (filter even? (range->list r1)))
            (range->list
             (range-filter-map (lambda (k) (and (even? k) (add1 k)))
                               r1)))
      (test (map + (range->list r1) (range->list r2))
            (range->list (range-filter-map + r1 r2)))
      (test 0 (range-length
               (range-filter-map (lambda (x y z) x)
                                 r1 r2 (iota-range 0))))
      )
    (test-assert (raises-type-exn (range-filter-map #t (iota-range 0))))
    (test-assert (raises-type-exn (range-filter-map values 0.2)))
    (test-assert
     (raises-type-exn (range-filter-map values (iota-range 0) #t)))
    )

  (test-group "range-filter-map->list"
    (test '() (range-filter-map->list (constantly #t) (iota-range 0)))

    (test/random-ranges (r1 r2)
      (test '() (range-filter-map->list (constantly #f) r1))
      (test (range->list r1) (range-filter-map->list values r1))
      (test-assert (every (lambda (x) (eqv? x 'z))
                          (range-filter-map->list (constantly 'z) r1)))
      (test (range-length r1)
            (length (range-filter-map->list (constantly 'z) r1)))
      (test (map add1 (filter even? (range->list r1)))
            (range-filter-map->list (lambda (k) (and (even? k) (add1 k)))
                                    r1))
      (test (map + (range->list r1) (range->list r2))
            (range-filter-map->list + r1 r2))
      (test '() (range-filter-map->list (lambda (x y z) x)
                                        r1 r2 (iota-range 0)))
      )
    (test-assert
     (raises-type-exn (range-filter-map->list #t (iota-range 0))))
    (test-assert (raises-type-exn (range-filter-map->list values 0.2)))
    (test-assert
     (raises-type-exn (range-filter-map->list values (iota-range 0) #t)))
    )

  ;;; filter & remove

  (test-group "range-filter"
    (test 0 (range-length (range-filter always (iota-range 0))))

    (test/random-ranges (r)
      (test 0 (range-length (range-filter never r)))
      (test-assert (range=? = r (range-filter always r)))
      (test (filter even? (range->list r))
            (range->list (range-filter even? r)))
      (test '(z z z z z)
            (range->list
             (range-filter symbol?
                           (range-append r (range 5 (constantly 'z))))))
      (test-assert
       (range=? =
                r
                (range-filter
                 number?
                 (range-append r (range 5 (constantly 'z))))))
      )
    (test-assert (raises-type-exn (range-filter #t (iota-range 0))))
    (test-assert (raises-type-exn (range-filter values 0.2)))
    )

  (test-group "range-filter->list"
    (test '() (range-filter->list always (iota-range 0)))

    (test/random-ranges (r)
      (test '() (range-filter->list never r))
      (test (range->list r) (range-filter->list always r))
      (test (filter even? (range->list r)) (range-filter->list even? r))
      (test '(z z z z z)
            (range-filter->list
             symbol?
             (range-append r (range 5 (constantly 'z)))))
      (test (range->list r)
            (range-filter->list
             number?
             (range-append r (range 5 (constantly 'z)))))
      )
    (test-assert
     (raises-type-exn (range-filter->list #t (iota-range 0))))
    (test-assert (raises-type-exn (range-filter->list values 0.2)))
    )

  (test-group "range-remove"
    (test 0 (range-length (range-remove never (iota-range 0))))

    (test/random-ranges (r)
      (test 0 (range-length (range-remove always r)))
      (test-assert (range=? = r (range-remove never r)))
      (test (remove even? (range->list r))
            (range->list (range-remove even? r)))
      (test '(z z z z z)
            (range->list
             (range-remove number?
                           (range-append r (range 5 (constantly 'z))))))
      (test-assert
       (range=? =
                r
                (range-remove
                 symbol?
                 (range-append r (range 5 (constantly 'z))))))
      )
    (test-assert (raises-type-exn (range-remove #t (iota-range 0))))
    (test-assert (raises-type-exn (range-remove values 0.2)))
    )

  (test-group "range-remove->list"
    (test '() (range-remove->list never (iota-range 0)))

    (test/random-ranges (r)
      (test '() (range-remove->list always r))
      (test (range->list r) (range-remove->list never r))
      (test (remove even? (range->list r)) (range-remove->list even? r))
      (test '(z z z z z)
            (range-remove->list
             number?
             (range-append r (range 5 (constantly 'z)))))
      (test (range->list r)
            (range-remove->list
             symbol?
             (range-append r (range 5 (constantly 'z)))))
      )
    (test-assert
     (raises-type-exn (range-remove->list #t (iota-range 0))))
    (test-assert (raises-type-exn (range-remove->list values 0.2)))
    )

  (test-group "range-fold"
    (test 'z (range-fold + 'z (iota-range 0)))

    (test/random-ranges (r1 r2)
      (when (positive? (range-length r1))
        (test (range-last r1) (range-fold (lambda (_acc x) x) #f r1)))
      (test (fold + 0 (range->list r1)) (range-fold + 0 r1))
      (test (reverse (range->list r1)) (range-fold xcons '() r1))
      ;; fold & range-fold have the same behavior on short lists
      ;; or ranges.
      ;; It's too bad they can't agree on argument order!
      (let ((f (lambda (ks a b) (cons (min a b) ks))))
        (test (fold (lambda (a b ks) (cons (min a b) ks))
                    '()
                    (range->list r1)
                    (range->list r2))
              (range-fold f '() r1 r2))
        (test '() (range-fold f '() r1 r2 (iota-range 0))))
      )

    (test-assert (raises-type-exn (range-fold #t 0 (iota-range 0))))
    (test-assert (raises-type-exn (range-fold + 0 0.5)))
    (test-assert (raises-type-exn (range-fold + 0 (iota-range 0) 't)))
    )

  (test-group "range-fold-right"
    (test 'z (range-fold-right + 'z (iota-range 0)))

    (test/random-ranges (r1 r2)
      (when (positive? (range-length r1))
        (test (range-first r1)
              (range-fold-right (lambda (_acc x) x) #f r1)))
      (test (fold-right + 0 (range->list r1)) (range-fold-right + 0 r1))
      (test (range->list r1) (range-fold-right xcons '() r1))
      ;; fold-right & range-fold-right have the same behavior on short
      ;; lists or ranges.
      ;; It's too bad they can't agree on argument order!
      (let ((f (lambda (ks a b) (cons (min a b) ks))))
        (test (fold-right (lambda (a b ks) (cons (min a b) ks))
                          '()
                          (range->list r1)
                          (range->list r2))
              (range-fold-right f '() r1 r2))
        (test '() (range-fold-right f '() r1 r2 (iota-range 0))))
      )

    (test-assert
     (raises-type-exn (range-fold-right #t 0 (iota-range 0))))
    (test-assert (raises-type-exn (range-fold-right + 0 0.5)))
    (test-assert
     (raises-type-exn (range-fold-right + 0 (iota-range 0) 't)))
    )
)

;;;; Searching

(test-group "Searching"
  (test-group "range-index"
    (test #f (range-index always (iota-range 0)))

    (test/random-ranges (r1 r2)
      (let ((lis1 (range->list r1)))
        ;; False if r1 is empty, and zero otherwise.
        (test (list-index always lis1) (range-index always r1))
        (test #f (range-index never r1))
        (test (list-index odd? lis1) (range-index odd? r1))
        (test (list-index even? lis1) (range-index even? r1))
        (test (list-index (lambda (k) (> k 1000)) lis1)
              (range-index (lambda (k) (> k 1000)) r1))
        (test (list-index < lis1 (range->list r2))
              (range-index < r1 r2))
        )
      (test #f (range-index always r1 (iota-range 0)))
      )

    (test-assert (raises-type-exn (range-index #t (iota-range 0))))
    (test-assert (raises-type-exn (range-index odd? 0.3)))
    (test-assert (raises-type-exn (range-index odd? (iota-range 0) #t)))
    )

  (test-group "range-index-right"
    (test #f (range-index-right always (iota-range 0)))
    (test 3 (range-index-right (lambda (x) (= x 3)) (iota-range 5)))
    (test 4 (range-index-right (lambda (x y) (< (+ x y) 30))
                               test-num-range
                               test-num-range))

    ;; Since there's no list-index-right, we test against
    ;; vector-index-right instead.
    (test/random-ranges (r1)
      (let ((vec1 (range->vector r1)))
        ;; False if r1 is empty, and the last index of r1 otherwise.
        (test (vector-index-right always vec1)
              (range-index-right always r1))
        (test #f (range-index-right never r1))
        (test (vector-index-right odd? vec1) (range-index-right odd? r1))
        (test (vector-index-right even? vec1) (range-index-right even? r1))
        (test (vector-index-right (lambda (k) (> k 1000)) vec1)
              (range-index-right (lambda (k) (> k 1000)) r1))
        (let* ((r2 (iota-range (range-length r1) 1000 10))
               (vec2 (range->vector r2))
               (divides? (lambda (m n) (zero? (quotient m n)))))
          (test (vector-index-right divides? vec1 vec2)
                (range-index-right divides? r1 r2))
          (test (vector-index-right divides? vec2 vec1)
                (range-index-right divides? r2 r1))
          ))
      (test-error (range-index-right always r (iota-range 0)))
      )

    (test-assert (raises-type-exn (range-index-right #t (iota-range 0))))
    (test-assert (raises-type-exn (range-index-right odd? 0.3)))
    (test-assert
     (raises-type-exn (range-index-right odd? (iota-range 0) #t)))
    )

  (test-group "range-take-while"
    (test 0 (range-length (range-take-while always (iota-range 0))))
    (test-assert (range=? =
                          (numeric-range 10 20)
                          (range-take-while (lambda (k) (< k 20))
                                            (numeric-range 10 30))))

    (test/random-ranges (r)
      (test-assert (range=? = r (range-take-while always r)))
      (test 0 (range-length (range-take-while never r)))
      (test (take-while (lambda (k) (< k 10000)) (range->list r))
            (range->list (range-take-while (lambda (k) (< k 10000)) r)))
      )
    (test-assert (raises-type-exn (range-take-while #t (iota-range 0))))
    (test-assert (raises-type-exn (range-take-while odd? #t)))
    )

  (test-group "range-take-while-right"
    (test 0 (range-length (range-take-while-right always
                                                  (iota-range 0))))
    (test-assert
     (range=? =
              (numeric-range 20 30)
              (range-take-while-right (lambda (k) (>= k 20))
                                      (numeric-range 10 30))))

    (test/random-ranges (r)
      (test-assert (range=? = r (range-take-while-right always r)))
      (test 0 (range-length (range-take-while-right never r)))
      ;; Yeech.  But there is no other take-while-right to
      ;; test against.
      (test (reverse
             (take-while (lambda (k) (< k 10000))
                         (reverse (range->list r))))
            (range->list
             (range-take-while-right (lambda (k) (< k 10000)) r)))
      )
    (test-assert
     (raises-type-exn (range-take-while-right #t (iota-range 0))))
    (test-assert (raises-type-exn (range-take-while-right odd? #t)))
    )

  (test-group "range-drop-while"
    (test 0 (range-length (range-drop-while always (iota-range 0))))
    (test-assert (range=? =
                          (numeric-range 20 30)
                          (range-drop-while (lambda (k) (< k 20))
                                            (numeric-range 10 30))))

    (test/random-ranges (r)
      (test-assert (range=? = r (range-drop-while never r)))
      (test 0 (range-length (range-drop-while always r)))
      (test (drop-while (lambda (k) (< k 10000)) (range->list r))
            (range->list (range-drop-while (lambda (k) (< k 10000)) r)))
      )
    (test-assert (raises-type-exn (range-drop-while #t (iota-range 0))))
    (test-assert (raises-type-exn (range-drop-while odd? #t)))
    )

  (test-group "range-drop-while-right"
    (test 0 (range-length (range-drop-while-right always
                                                  (iota-range 0))))
    (test-assert
     (range=? =
              (numeric-range 10 20)
              (range-drop-while-right (lambda (k) (>= k 20))
                                      (numeric-range 10 30))))

    (test/random-ranges (r)
      (test-assert (range=? = r (range-drop-while-right never r)))
      (test 0 (range-length (range-drop-while-right always r)))
      ;; Yeech.  But there is no other drop-while-right to
      ;; test against.
      (test (reverse
             (drop-while (lambda (k) (< k 10000))
                         (reverse (range->list r))))
            (range->list
             (range-drop-while-right (lambda (k) (< k 10000)) r)))
      )
    (test-assert
     (raises-type-exn (range-drop-while-right #t (iota-range 0))))
    (test-assert (raises-type-exn (range-drop-while-right odd? #t)))
    )

  (test-group "range-take-while ++ range-drop-while"
    (test/random-ranges (r)
      (let ((preds (list odd?
                         even?
                         (lambda (k) (< k 10000))
                         (lambda (k) (zero? (modulo k 10))))))
        (test-assert
         (every (lambda (p)
                  (range=? =
                           r
                           (range-append (range-take-while p r)
                                         (range-drop-while p r))))
                preds)))
      ))

  (test-group "range-drop-while-right ++ range-take-while-right"
    (test/random-ranges (r)
      (let ((preds (list odd?
                         even?
                         (lambda (k) (< k 10000))
                         (lambda (k) (zero? (modulo k 10))))))
        (test-assert
         (every (lambda (p)
                  (range=? =
                           r
                           (range-append (range-drop-while-right p r)
                                         (range-take-while-right p r))))
                preds)))
      ))
)

(test-exit)
